from __future__ import unicode_literals, division, print_function, absolute_import
from builtins import range, object

import warnings
from collections import Counter
from joblib import Parallel, delayed
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler, OneHotEncoder
import sklearn.linear_model as lm
from sympy.utilities.autowrap import ufuncify
import pint

from .feateng import generate_features, n_cols_generated
from .featsel import select_features


def _parse_units(units, ureg=None):
    """
    Convert a dict with string units to pint quantities.

    Inputs:
        - units: dict with {"variable_name": "unit"}
        - ureg: optional: a pint UnitRegistry

    Returns
        - parsed_units: dict with {"variable_name": pint Quantity}
    """
    if ureg is None:
        ureg = pint.UnitRegistry(auto_reduce_dimensions=True, autoconvert_offset_to_baseunit=True)
    parsed_units = {}
    for c in units:
        try:
            parsed_units[c] = ureg.parse_expression(units[c])
        except pint.UndefinedUnitError as e:
            print("[AutoFeatRegression] WARNING: unit %r of column %r was not recognized and will be ignored!" % (units[c], c))
            parsed_units[c] = ureg.parse_expression("")
        parsed_units[c].__dict__["_magnitude"] = 1.
    return parsed_units


class AutoFeatRegression(object):

    def __init__(
        self,
        categorical_cols=[],
        feateng_cols=None,
        units={},
        feateng_steps=3,
        featsel_runs=5,
        max_gb=None,
        transformations=["exp", "log", "abs", "sqrt", "^2", "^3", "1/"],
        n_jobs=1,
    ):
        """
        multi-step feature engineering and cross-validated feature selection to generate promising additional
        features for your dataset and train a Lasso regression model with them.

        Inputs:
            - categorical_cols: list of column names of categorical features; these will be transformed into
                                0/1 encoding and not used in the feature engineering part (default: [])
            - feateng_cols: list of column names that should be used for the feature engineering part
                            (default None --> all except categorical_cols)
            - units: dictionary with {col_name: unit} where unit is a string that can be converted into a pint unit.
                     all columns without units are dimensionless and can be combined with any other column
                     (default: {} --> all columns are dimensionless).
            - feateng_steps: number of steps to perform in the feature engineering part (int; default: 3)
            - featsel_runs: number of times to perform in the feature selection part with a random fraction of data points (int; default: 5)
            - max_gb: if an int is given: maximum number of gigabytes to use in the process (i.e. mostly the
                      feature engineering part). this is no guarantee! it will lead to subsampling of the
                      data points if the new dataframe generated is n_rows * n_cols * 32bit > max_gb
            - transformations: list of transformations that should be applied; possible elements:
                               "exp", "log", "abs", "sqrt", "^2", "^3", "1/", "1+", "1-", "sin", "cos", "exp-", "2^"
                               (first 7, i.e., up to 1/, are applied by default)
            - n_jobs: how many jobs to run when selecting the features in parallel (default: 1)

        Note: when giving categorical_cols or feateng_cols, X later (i.e. when calling fit/fit_transform) has to be a DataFrame
        """
        self.categorical_cols = categorical_cols
        self.feateng_cols = feateng_cols
        self.units = units
        self.feateng_steps = feateng_steps
        self.max_gb = max_gb
        self.featsel_runs = featsel_runs
        self.transformations = transformations
        self.n_jobs = n_jobs
        # sympy formulas to generate new features
        self.feature_formulas = {}
        # compiled feature functions with columns
        self.feature_functions = {}
        # list of good new features that should be generated when calling transform()
        self.new_feat_cols = []
        # trained regression model
        self.regression_model = None

    def __getstate__(self):
        """
        get dict for pickling without feature_functions as they are not pickleable
        """
        return {k: self.__dict__[k] if k != "feature_functions" else {} for k in self.__dict__}

    def _transform_categorical_cols(self, df, cols={}):
        """
        Transform categorical features into 0/1 encoding.

        Inputs:
            - df: pandas dataframe with original features
            - cols: set of columns, from which the categorical_cols will be removed (optional)
        Returns:
            - df: dataframe with categorical features transformed into multiple 0/1 columns
            - if cols was given: updated cols set
        """
        if self.categorical_cols:
            e = OneHotEncoder(sparse=False, categories='auto')
            for c in self.categorical_cols:
                if cols:
                    cols.remove(c)
                ohe = e.fit_transform(df[c].values[:, None])
                df = df.join(pd.DataFrame(ohe, columns=["%s_%r" % (str(c), i) for i in e.categories_[0]], index=df.index))
            # remove the categorical column from our columns to consider
            df.drop(columns=self.categorical_cols, inplace=True)
        if cols:
            return df, cols
        return df

    def _apply_pi_theorem(self, df):
        if self.units:
            ureg = pint.UnitRegistry(auto_reduce_dimensions=True, autoconvert_offset_to_baseunit=True)
            parsed_units = _parse_units(self.units, ureg)
            # use only original features
            parsed_units = {c: parsed_units[c] for c in self.feateng_cols if not parsed_units[c].dimensionless}
            print("[AutoFeatRegression] Applying the Pi Theorem")
            pi_theorem_results = ureg.pi_theorem(parsed_units)
            for i, r in enumerate(pi_theorem_results, 1):
                print("Pi Theorem %i: " % i, pint.formatter(r.items()))
                # compute the final result by multiplying and taking the power of
                cols = sorted(r)
                ptr = df[cols[0]].values**r[cols[0]]
                for c in cols[1:]:
                    ptr *= df[c].values**r[c]
                df["PT%i: %s" % (i, pint.formatter(r.items()).replace(" ", ""))] = ptr
        return df

    def _generate_features(self, df, new_feat_cols):
        """
        Generate additional features based on the feature formulas for all data points in the df.
        Only works after the model was fitted.

        Inputs:
            - df: pandas dataframe with original features
            - new_feat_cols: names of new features that should be generated (keys of self.feature_formulas)
        Returns:
            - df: dataframe with the additional feature columns added
        """
        assert new_feat_cols[0] in self.feature_formulas,\
            "[AutoFeatRegression] First call fit or fit_transform to generate the features!"
        print("[AutoFeatRegression] Computing %i new features." % len(new_feat_cols))
        # generate all good feature; unscaled this time
        feat_array = np.zeros((len(df), len(new_feat_cols)))
        for i, expr in enumerate(new_feat_cols):
            print("[AutoFeatRegression] %5i/%5i" % (i, len(new_feat_cols)), end="\r")
            if expr not in self.feature_functions:
                # generate a substitution expression based on all the original symbols of the original features
                # for the given generated feature in good cols
                # since sympy can handle only up to 32 original features in ufunctify, we need to check which features
                # to consider here, therefore perform some crude check to limit the number of features used
                cols = [c for c in self.feateng_cols if c in expr]
                try:
                    f = ufuncify((self.feature_formulas[c] for c in cols), self.feature_formulas[expr])
                except:
                    print("[AutoFeatRegression] Error while processing expression: %r" % expr)
                    raise
                self.feature_functions[expr] = (cols, f)
            else:
                cols, f = self.feature_functions[expr]
            try:
                feat_array[:, i] = f(*(df[c].values for c in cols))
            except RuntimeWarning as e:
                print("[AutoFeatRegression] Problem while evaluating expression: %r with columns %r - are maybe some values 0 that shouldn't be?" % (expr, cols))
                raise
        print("[AutoFeatRegression] %5i/%5i ...done." % (len(new_feat_cols), len(new_feat_cols)))
        df = df.join(pd.DataFrame(feat_array, columns=new_feat_cols, index=df.index))
        return df

    def fit_transform(self, X, y):
        """
        Fits the regression model and returns a new dataframe with the additional features.

        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
            - y: pandas dataframe or numpy array with one target variable for all n_datapoints
        Returns:
            - new_df: new pandas dataframe with all the original features (except categorical features transformed
                      into multiple 0/1 columns) and the most promising engineered features. This df can then be
                      used to train your final model.

        Please ensure that X only contains valid feature columns (including possible categorical variables).

        Note: we strongly encourage you to name your features X1 ...  Xn or something simple like this before passing
              a DataFrame to this model. This can help avoid potential problems with sympy later on.
        """
        # we need y as a numpy array later anyways
        if isinstance(y, pd.DataFrame) or isinstance(y, pd.core.series.Series):
            target = y.values
        else:
            target = y
        if len(target.shape) > 1:
            target = target[:, 0]
        # possibly transform X to a dataframe or copy to mess around with it
        if not isinstance(X, pd.DataFrame):
            df = pd.DataFrame(X, columns=["x%i" % i for i in range(10, 10 + X.shape[1])])
        else:
            df = X.copy()
            # in case X was created by filtering out some rows
            df.reset_index(drop=True, inplace=True)
        # possibly convert categorical columns
        cols = set(df.columns)
        df, cols = self._transform_categorical_cols(df, cols)
        # if we're not given specific feateng_cols, then just take all columns except categorical
        if not self.feateng_cols:
            self.feateng_cols = sorted(cols)
        # convert units to proper pint units
        if self.units:
            # need units for only and all feateng columns
            self.units = {c: self.units[c] if c in self.units else "" for c in self.feateng_cols}
            # apply pi-theorem -- additional columns are not used for regular feature engineering (for now)!
            df = self._apply_pi_theorem(df)
        # subsample data points and targets in case we'll generate too many features
        # (n_rows * n_cols * 32/8)/1000000000 <= max_gb
        n_cols = n_cols_generated(len(self.feateng_cols), self.feateng_steps, len(self.transformations))
        n_gb = (len(df) * n_cols) / 250000000
        print("[AutoFeatRegression] The %i step feature engineering process could generate up to %i features." % (self.feateng_steps, n_cols))
        print("[AutoFeatRegression] With %i data points this new feature matrix would use about %.2f gb of space." % (len(df), n_gb))
        if self.max_gb and n_gb > self.max_gb:
            n_rows = int(self.max_gb * 250000000 / n_cols)
            print("[AutoFeatRegression] As you specified a limit of %.1d gb, the number of data points is subsampled to %i" % (self.max_gb, n_rows))
            subsample_idx = np.random.permutation(list(df.index))[:n_rows]
            df_subs = df.iloc[subsample_idx]
            df_subs.reset_index(drop=True, inplace=True)
            target_sub = target[subsample_idx]
        else:
            df_subs = df
            target_sub = target
        # generate features and scale to have 0 mean and unit std
        df_scaled, self.feature_formulas = generate_features(df_subs, self.feateng_cols, _parse_units(self.units),
                                                             self.feateng_steps, self.transformations)
        s = StandardScaler()
        df_scaled = pd.DataFrame(s.fit_transform(df_scaled), columns=df_scaled.columns, dtype=np.float32)
        # do sort of a cross-validation (i.e., randomly subsample data points)
        # to select features with high relevance
        selected_columns = []
        idx = list(df_scaled.index)
        print("[AutoFeatRegression] Selecting good features in %i runs" % self.featsel_runs)

        # select good features in 5 runs in parallel
        def run_select_features(i):
            np.random.seed(i)
            train_idx = np.random.permutation(idx)[:int(0.8 * len(idx))]
            return select_features(df_scaled.iloc[train_idx], target_sub[train_idx], True, eps=1e-8)
        if self.n_jobs == 1:
            # only use parallelization code if you actually parallelize
            selected_columns = []
            for i in range(self.featsel_runs):
                selected_columns.extend(run_select_features(i))
        else:
            flatten_lists = lambda l: [item for sublist in l for item in sublist]
            selected_columns = flatten_lists(Parallel(n_jobs=self.n_jobs, verbose=100)(delayed(run_select_features)(i) for i in range(self.featsel_runs)))
        # check in how many runs each feature was selected and only takes those that were selected in more than one run
        selected_columns = Counter(selected_columns)
        original_features = list(df.columns)
        good_cols = [c for c in selected_columns if selected_columns[c] > 1 and c not in original_features]
        print("[AutoFeatRegression] %i features occurred in more than one featsel run." % len(good_cols))
        # train another regression model on these features
        df_scaled = df_scaled[original_features + good_cols]
        X = df_scaled.values
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            reg = lm.LassoLarsCV(eps=1e-8)
            reg.fit(X, target_sub)
        weights = dict(zip(list(df_scaled.columns), reg.coef_))
        good_cols = [c for c in weights if abs(weights[c]) >= 1e-5 and c not in original_features]
        print("[AutoFeatRegression] %i new features selected." % len(good_cols))
        # re-generate all good feature again; unscaled this time
        df = self._generate_features(df, good_cols)
        print("[AutoFeatRegression] Training final regression model.")
        # train final regression model all selected features (twice)
        for i in range(2):
            X = df.values
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                reg = lm.LassoLarsCV(eps=1e-16)
                reg.fit(X, target)
            weights = dict(zip(list(df.columns), reg.coef_))
            if not i:
                self.new_feat_cols = [c for c in weights if abs(weights[c] * df[c].std()) >= 1e-6 and c not in original_features]
                df = df[original_features + self.new_feat_cols]
        # filter out unnecessary junk from self.feature_formulas
        self.feature_formulas = {f: self.feature_formulas[f] for f in self.new_feat_cols + self.feateng_cols}
        self.feature_functions = {f: self.feature_functions[f] for f in self.new_feat_cols}
        print("[AutoFeatRegression] Trained model coefficients:")
        print(reg.intercept_)
        for c in sorted(weights, key=lambda x: abs(weights[x]), reverse=True):
            if abs(weights[c]) < 1e-5:
                break
            print("%.6f * %s" % (weights[c], c))
        print("[AutoFeatRegression] Final R^2: %.4f" % reg.score(X, target))
        self.regression_model = reg
        # final dataframe contains original columns, good additional columns, and target column
        print("[AutoFeatRegression] Final dataframe with %i feature columns (%i new)." % (len(df.columns), len(df.columns) - len(original_features)))
        return df

    def fit(self, X, y):
        print("[AutoFeatRegression] Warning: This just calls fit_transform() but does not return the transformed dataframe.")
        print("[AutoFeatRegression] It is much more efficient to call fit_transform() instead of fit() and transform()!")
        _ = self.fit_transform(X, y)

    def transform(self, X):
        """
        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
        Returns:
            - new_df: new pandas dataframe with all the original features (except categorical features transformed
                      into multiple 0/1 columns) and the most promising engineered features. This df can then be
                      used to train your final model.
        """
        assert self.regression_model is not None, "[AutoFeatRegression] First call fit or fit_transform to train the model!"
        # possibly transform X to a dataframe or copy to mess around with it
        if not isinstance(X, pd.DataFrame):
            df = pd.DataFrame(X, columns=["x%i" % i for i in range(10, 10 + X.shape[1])])
        else:
            df = X.copy()
        # possibly convert categorical columns
        df = self._transform_categorical_cols(df)
        # possibly apply pi-theorem
        df = self._apply_pi_theorem(df)
        # generate engineered features
        df = self._generate_features(df, self.new_feat_cols)
        return df

    def predict(self, X):
        """
        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
        Returns:
            - y_pred: vector of predicted targets return by regression_model.predict()
        """
        assert self.regression_model is not None, "[AutoFeatRegression] First call fit or fit_transform to train the model!"
        # possibly transform X to a dataframe or copy to mess around with it
        if not isinstance(X, pd.DataFrame):
            df = pd.DataFrame(X, columns=["x%i" % i for i in range(10, 10 + X.shape[1])])
        else:
            df = X
        # check if the dataframe was already transformed
        if not self.new_feat_cols[0] in df:
            df = self.transform(df)
        return self.regression_model.predict(df.values)

    def score(self, X, y):
        """
        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
            - y: pandas dataframe or numpy array with one target variable for all n_datapoints
        Returns:
            - R^2 value returned by regression_model.score()
        """
        assert self.regression_model is not None, "[AutoFeatRegression] First call fit or fit_transform to train the model!"
        # possibly transform X to a dataframe or copy to mess around with it
        if not isinstance(X, pd.DataFrame):
            df = pd.DataFrame(X, columns=["x%i" % i for i in range(10, 10 + X.shape[1])])
        else:
            df = X
        # check if the dataframe was already transformed
        if not self.new_feat_cols[0] in df:
            df = self.transform(df)
        # get the target as an array
        if isinstance(y, pd.DataFrame) or isinstance(y, pd.core.series.Series):
            y = y.values
        return self.regression_model.score(df.values, y)



"""
Setup script for autofeat based on the below example:
A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject

upload new versions:
- increase version number our you'll get complaints
- delete old folders or at least contents of dist/
- python setup.py sdist
- python setup.py bdist_wheel
- twine upload dist/*
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='autofeat',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1',

    description='Automatic Feature Engineering and Selection Linear Regression Model',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/cod3licious/autofeat',

    # Author details
    author='Franziska Horn',
    author_email='cod3licious@gmail.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # What does your project relate to?
    keywords='feature engineering feature selection linear regression',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    # you can also use
    # packages=[''],
    # to include all files in the root directory, however then when they are installed somewhere it's a bit messy
    packages=find_packages(exclude=['autofeat_test.ipynb']),

    # Alternatively, if you want to distribute just a my_module.py (really only single files with unique names), uncomment
    # this:
    # py_modules=["ml_utils", "dict_utils"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['future', 'numpy', 'pandas', 'sympy', 'scikit-learn', 'pint'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },
)



from __future__ import unicode_literals, division, print_function, absolute_import
from builtins import str
import operator as op
from functools import reduce
from itertools import combinations, product
import numpy as np
import pandas as pd
import sympy
from sympy.utilities.autowrap import ufuncify
import pint


def ncr(n, r):
    # compute number of combinations for n chose r
    r = min(r, n - r)
    numer = reduce(op.mul, range(n, n - r, -1), 1)
    denom = reduce(op.mul, range(1, r + 1), 1)
    return numer // denom


def n_cols_generated(n_features, max_steps, n_transformations=7, n_combinations=4):
    """
    computes the upper bound of how many features will be generated based on n_features to start with
    and max_steps feateng steps.
    """
    # n_transformations is 1-len(func_transform) because either abs() or sqrt and log will be applied
    n_transformations -= 1
    original_cols = n_features
    new_cols = 0
    new_new_cols = 0
    # count additions at the highest level
    n_additions = 0
    steps = 1
    if steps <= max_steps:
        # Step 1: apply transformations to original features
        original_cols += n_features * n_transformations
        n_additions += n_features * 2
        steps += 1
    if steps <= max_steps:
        # Step 2: first combination of features
        new_cols = n_combinations * (ncr(original_cols, 2))
        n_additions += 3 * new_cols // 4
        steps += 1
    while steps <= max_steps:
        # apply transformations on these new features
        n_additions += new_cols * 2
        new_cols += new_cols * n_transformations
        steps += 1
        # get combinations of old and new features
        if steps <= max_steps:
            new_new_cols = n_combinations * (original_cols * new_cols)
            n_additions += 3 * new_new_cols // 4
            steps += 1
        # and combinations of new features within themselves
        if steps <= max_steps:
            n = n_combinations * (ncr(new_cols, 2))
            new_new_cols += n
            n_additions += 3 * n // 4
            steps += 1
            # update old and new features and repeat
            original_cols += new_cols
            new_cols = new_new_cols
            new_new_cols = 0
    # finally, apply transformation on the last new features
    if steps <= max_steps:
        n_additions += new_cols * 2
        new_cols += new_cols * n_transformations
    return original_cols + new_cols + new_new_cols - n_additions


def generate_features(
    df_org,
    start_features=[],
    units={},
    max_steps=3,
    transformations=["exp", "log", "abs", "sqrt", "^2", "^3", "1/"],
):
    """
    Given a DataFrame with original features, perform the feature engineering routine for max_steps.
    It starts with a transformation of the original features (applying log, ^2, sqrt, etc.),
    then in the next step, the features are combined (x+y, x*y, ...), and in further steps, the resulting
    features are again transformed and combinations of the resulting features are computed.

    Inputs:
        - df_org: pandas DataFrame with original features in columns
        - start_features: list with column names for df_org with features that should be considered for expansion
                          (default: [] --> all columns)
        - units: a dict with {column_name: pint.Quantity}: some operations like x+y can only be performed if the
                 features have comparable units (default {}: all combinations are allowed)
                 careful: will be modified in place!
        - max_steps: how many feature engineering steps should be performed. Default is 4, this produces:
            Step 1: transformation of original features
            Step 2: first combination of features
            Step 3: transformation of new features
            Step 4: combination of old and new features
            --> with 3 original features you will already end up with around 200k features after this!
        - transformations: list of transformations that should be applied; possible elements:
                           "exp", "log", "abs", "sqrt", "^2", "^3", "1/", "1+", "1-", "sin", "cos", "exp-", "2^"
                           (first 7, i.e., up to 1/, are applied by default)
    Returns:
        - df: new DataFrame with all features in columns
        - feature_pool: dict with {col: sympy formula} formulas to generate each feature
    """
    # initialize the feature pool with columns from the dataframe
    if not start_features:
        start_features = df_org.columns
    feature_pool = {c: sympy.symbols(c.replace(" ", ""), real=True) for c in start_features}
    if max_steps < 1:
        print("Warning: no features generated for max_steps < 1.")
        return df_org, feature_pool
    # get a copy of the dataframe - this is where all the features will be added
    df = pd.DataFrame(df_org.copy(), dtype=np.float32)

    def apply_tranformations(features_list):
        # feature transformations
        func_transform = {
            "exp": lambda x: sympy.exp(x),
            "exp-": lambda x: sympy.exp(-x),
            "log": lambda x: sympy.log(x),
            "abs": lambda x: sympy.Abs(x),
            "sqrt": lambda x: sympy.sqrt(x),
            "sin": lambda x: sympy.sin(x),
            "cos": lambda x: sympy.cos(x),
            "2^": lambda x: 2**x,
            "^2": lambda x: x**2,
            "^3": lambda x: x**3,
            "1+": lambda x: 1 + x,
            "1-": lambda x: 1 - x,
            "1/": lambda x: 1 / x
        }
        func_transform_units = {
            "exp": lambda x: np.exp(x),
            "exp-": lambda x: np.exp(-x),
            "log": lambda x: np.log(x),
            "abs": lambda x: np.abs(x),
            "sqrt": lambda x: np.sqrt(x),
            "sin": lambda x: np.sin(x),
            "cos": lambda x: np.cos(x),
            "2^": lambda x: np.exp(x),
            "^2": lambda x: x**2,
            "^3": lambda x: x**3,
            "1+": lambda x: 1 + x,
            "1-": lambda x: 1 - x,
            "1/": lambda x: 1 / x
        }
        # conditions on the original features that have to be met to apply the transformation
        func_transform_cond = {
            "exp": lambda x: np.all(x < 10),
            "exp-": lambda x: np.all(-x < 10),
            "log": lambda x: np.all(x > 0),
            "abs": lambda x: np.any(x < 0),
            "sqrt": lambda x: np.all(x >= 0),
            "sin": lambda x: True,
            "cos": lambda x: True,
            "2^": lambda x: np.all(x < 50),
            "^2": lambda x: np.all(np.abs(x) < 1000000),
            "^3": lambda x: np.all(np.abs(x) < 10000),
            "1+": lambda x: True,
            "1-": lambda x: True,
            "1/": lambda x: np.all(x != 0)
        }
        # apply transformations to the features in the given features list
        # modifies global variables df and feature_pool!
        nonlocal df, feature_pool, units
        # returns a list of new features that were generated
        new_features = []
        # store all new features in a preallocated numpy array before adding it to the dataframe
        feat_array = np.zeros((df.shape[0], len(features_list) * len(transformations)), dtype=np.float32)
        for i, feat in enumerate(features_list):
            if not i % 100:
                print("%15i/%15i" % (i, len(features_list)), end="\r")
            for ft in transformations:
                # check if transformation is valid for particular feature (i.e. given actual numerical values)
                if func_transform_cond[ft](df[feat]):
                    # get the expression (based on the primary features)
                    expr = func_transform[ft](feature_pool[feat])
                    expr_name = str(expr)
                    # we're simplifying expressions, so we might already that one
                    if expr_name not in feature_pool:
                        # if we're given units, check if the operation is legal
                        if units:
                            try:
                                units[expr_name] = func_transform_units[ft](units[feat])
                                units[expr_name].__dict__["_magnitude"] = 1.
                            except (pint.DimensionalityError, pint.OffsetUnitCalculusError) as e:
                                continue
                        feature_pool[expr_name] = expr
                        # create temporary variable expression and apply it to precomputed feature
                        t = sympy.symbols("t")
                        expr_temp = func_transform[ft](t)
                        f = ufuncify(t, expr_temp)
                        new_feat = np.array(f(df[feat].values), dtype=np.float32)
                        if np.isfinite(new_feat).all():
                            feat_array[:, len(new_features)] = new_feat
                            new_features.append(expr_name)
        print("%15i/%15i done." % (len(new_features), len(features_list)))
        df = df.join(pd.DataFrame(feat_array[:, :len(new_features)], columns=new_features, index=df.index, dtype=np.float32))
        return new_features

    def get_feature_combinations(feature_tuples):
        # new features as combinations of two other features
        func_combinations = {
            "x+y": lambda x, y: x + y,
            "x*y": lambda x, y: x * y,
            "x-y": lambda x, y: x - y,
            "y-x": lambda x, y: y - x
        }
        # get all feature combinations for the given feature tuples
        # modifies global variables df and feature_pool!
        nonlocal df, feature_pool, units
        # returns a list of new features that were generated
        new_features = []
        # store all new features in a preallocated numpy array before adding it to the dataframe
        feat_array = np.zeros((df.shape[0], len(feature_tuples) * len(func_combinations)), dtype=np.float32)
        for i, (feat1, feat2) in enumerate(feature_tuples):
            if not i % 100:
                print("%15i/%15i" % (i, len(feature_tuples)), end="\r")
            for fc in func_combinations:
                expr = func_combinations[fc](feature_pool[feat1], feature_pool[feat2])
                expr_name = str(expr)
                if expr_name not in feature_pool:
                    # if we're given units, check if the operation is legal
                    if units:
                        try:
                            units[expr_name] = func_combinations[fc](units[feat1], units[feat2])
                            units[expr_name].__dict__["_magnitude"] = 1.
                        except (pint.DimensionalityError, pint.OffsetUnitCalculusError) as e:
                            continue
                    feature_pool[expr_name] = expr
                    # create temporary variable expression to apply it to precomputed features
                    s, t = sympy.symbols("s t")
                    expr_temp = func_combinations[fc](s, t)
                    f = ufuncify((s, t), expr_temp)
                    new_feat = np.array(f(df[feat1].values, df[feat2].values), dtype=np.float32)
                    if np.isfinite(new_feat).all():
                        feat_array[:, len(new_features)] = new_feat
                        new_features.append(expr_name)
        print("%15i/%15i done." % (len(new_features), len(feature_tuples)))
        df = df.join(pd.DataFrame(feat_array[:, :len(new_features)], columns=new_features, index=df.index, dtype=np.float32))
        return new_features

    # get transformations of initial features
    steps = 1
    print("Step 1: transformation of original features")
    original_features = list(feature_pool.keys())
    original_features.extend(apply_tranformations(original_features))
    steps += 1
    # get combinations of first feature set
    if steps <= max_steps:
        print("Step 2: first combination of features")
        new_features = get_feature_combinations(list(combinations(original_features, 2)))
        steps += 1
    while steps <= max_steps:
        # apply transformations on these new features
        print("Step %i: transformation of new features" % steps)
        new_features.extend(apply_tranformations(new_features))
        steps += 1
        # get combinations of old and new features
        if steps <= max_steps:
            print("Step %i: combination of old and new features" % steps)
            new_new_features = get_feature_combinations(list(product(original_features, new_features)))
            steps += 1
        # and combinations of new features within themselves
        if steps <= max_steps:
            print("Step %i: combination of new features" % steps)
            new_new_features.extend(get_feature_combinations(list(combinations(new_features, 2))))
            steps += 1
            # update old and new features and repeat
            original_features.extend(new_features)
            new_features = new_new_features
    # finally, apply transformation on the last new features
    if steps <= max_steps:
        print("Step %i: transformation of last new features" % steps)
        new_features.extend(apply_tranformations(new_features))

    # sort out all features that are just additions on the highest level
    cols = [c for c in list(df.columns) if not (c in feature_pool and feature_pool[c].func == sympy.add.Add)]
    return df[cols], feature_pool



from __future__ import unicode_literals, division, print_function, absolute_import
from builtins import zip
import warnings
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
import sklearn.linear_model as lm


def select_features(df, target, df_scaled=False, max_it=100, eps=1e-16):
    """
    Inputs:
        - df: nxp pandas DataFrame with n data points and p features; to avoid overfitting, only provide data belonging
              to the n training data points. The variables should be scaled to have 0 mean and unit variance. If this is
              not the case, set df_scaled to False and it will be done for you.
        - target: n dimensional array with targets corresponding to the data points in df
        - df_scaled: (bool) whether df is already scaled to have 0 mean and unit variance (default: False)
        - max_it: how many iterations will be performed at most
        - eps: eps parameter for LassoLarsCV regression model (might need to increase that to ~1e-8 or 1e-5 if you get a warning)
    Returns:
        - good_cols: list of column names for df with which a regression model can be trained
    """
    s = StandardScaler()
    # for performance reasons the scaled data can already be given
    if not df_scaled:
        # scale features to have 0 mean and unit std
        print("scaling data...", end="")
        df = pd.DataFrame(s.fit_transform(df), columns=df.columns, dtype=np.float32)
        print("done.")

    # good cols contains the currently considered good features (=columns)
    good_cols = []
    best_cols = []
    # we want to select up to thr features (how much a regression model is comfortable with)
    thr = int(0.5 * df.shape[0])
    # our first target is the original target variable; later we operate on target - predicted_target
    new_target = target
    residual = np.mean(np.abs(target))
    last_residuals = np.zeros(max_it)
    smallest_residual = 10. * residual
    it = 0
    # we try optimizing features until we have converged or run over max_it
    while (it < max_it) and (not np.sum(np.isclose(residual, last_residuals)) >= 2):
        if not it % 10:
            print("iteration %3i; %3i good cols with residual: %.6f" % (it, len(good_cols), residual))
        last_residuals[it] = residual
        it += 1
        # select new possibly good columns from all but the currently considered good columns
        cols = set(df.columns)
        cols.difference_update(good_cols)
        cols_list = list(cols)
        # compute the absolute correlation of the (scaled) features with the (scaled) target variable
        w = np.abs(np.dot(s.fit_transform(new_target[:, None])[:, 0], df[cols_list].values))
        # add promising features such that len(previous good cols + new cols) = thr
        good_cols.extend([cols_list[c] for c in np.argsort(w)[-(thr - len(good_cols)):]])
        # compute the regression residual based on the best features so far
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            reg = lm.LassoLarsCV(eps=eps)
            X = df[good_cols].values
            reg.fit(X, target)
        new_target = target - reg.predict(X)
        residual = np.mean(np.abs(new_target))
        # update the good columns based on the regression coefficients
        weights = dict(zip(good_cols, reg.coef_))
        good_cols = [c for c in weights if abs(weights[c]) > 1e-6]
        if residual < smallest_residual:
            smallest_residual = residual
            best_cols = [c for c in good_cols]
    print("iteration %3i; %3i good cols with residual: %.6f  --> done." % (it, len(best_cols), smallest_residual))
    return best_cols
